/* -*- knf -*- */
/* $AMSEL: grammar_de.y,v 1.3 2004/05/23 15:03:38 reyk Exp $ */

/*
 * Copyright (c) 2004 .vantronix | secure systems GmbH
 * Copyright (c) 2004 SSV Embedded Systems GmbH
 *
 * All rights reserved.
 *
 * Written by Reyk Floeter <reyk@vantronix.net>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  *
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/* 
 * amwall German language version: Please use grammar_en.y
 * (pseudo-english)! This is just an example of porting the amwall
 * grammar/vocabulary to other native languages.  It's not as powerful
 * as the englisch version since it's rarely developed. Nevertheless,
 * someday I'll finish this version and maybe add some other language
 * modules, too...
 *
 * English:
 * "accept any from iface eth0 to iface eth1;"
 * German:
 * "Akzeptiere alles von der Schnittstelle eth0 zu der Schnittstelle eth1;"
 * 1337:
 * "411 y0uR 94ck375 fR0m n1c eth0 70 n1c eth1 4R3 b310ng 70 u5;"
 */

%{
#include "main.h"
#include "tables.h"
#include "ipvs.h"
#include "parser.h"
#include "grammar.h"

FILE *fin = NULL;
int lineno = 1;
int errors = 0;
char *infile;

static int ret;
static enum amselwall_table_type table_type;
static enum amselwall_entry_type entry_type;
static struct amselwall_table *table = NULL;
static struct amselwall_chain *chain = NULL;
static struct amselwall_rule *rule = NULL;
%}

%token UND DEM DER DIE DEN DAS ALLE ALLES VON ZU WENN MIT EIN VERWENDE
%token SETZE KETTE ETHERNET IPV4 IPV6 FILTER NAT MANGLE IRGENDWOHER IRGENDEINEM
%token EINGABE AUSGABE WEITERLEITUNG GRUNDSATZ IM FALLE WEISE AB BIS AUF ABER NUR
%token AKZEPTIERE VERWERFE PROTOKOLLIERE VERWERFEN AKZEPTIEREN FRAGMENTE KEINE
%token PROTOKOLL PORT QUELLE ZIEL SCHNITTSTELLE OPTIONEN VERSUCHE FUER
%token <v.string> STRING
%type <v.string> string source_l dest_l optstring policy
%type <v.number> negation negation_c
%%

/* 
 * Grammar
 */

grammar		: /* nothing */
                | grammar nl
                | grammar endl
                | grammar varset endl
		| grammar ipv4 endl
		| grammar ipv6 endl
		| grammar ethernet endl
		| grammar error endl 
{ 
	errors++; 
}
		;

ipv4            : IPV4
{
	table_type = AMSELWALL_TABLE_IPV4;
}
                  filter
                ;

ipv6            : IPV6
{
	yyerror("IPv6 support not yet implemented, use 'ip6tables'");
	YYERROR;
}
                ;

ethernet        : ETHERNET
{
	yyerror("Ethernet bridging support not yet implemented, use 'ebtables'");
	YYERROR;
}
                ;

/*
 * Misc types
 */

string		: string STRING				
{
	if (asprintf(&$$, "%s %s", $1, $2) == -1) {
		free($1);
		free($2);
		fatal("string: asprintf");
	}
	free($1);
	free($2);
}
		| STRING
		;

varset		: STRING '=' string		
{
	if (amwall_symset($1, $3, 0) == -1) {
		free($1);
		free($3);
		fatal("cannot store variable");
	}
	
	free($1);
	free($3);
}
		;

/*
 * IPv4 table "filter"
 */

filter          : FILTER optnl '{' optnl
{
	if((ret = amwall_add_table("filter", table_type, &table)) != 0) {
		yyerror("unable to add table 'filter': %s\n", strerror(-ret));
		YYERROR;
	}
}
                  filter_opts '}'
{
	table = NULL;
}
		;

filter_opts     : filter_opts filter_opts_l
                | filter_opts nl
                | filter_opts_l
                ;

filter_opts_l   : /* nothing */
                | KETTE string 
{ 
	if((ret = amwall_add_chain(table, $2, AMSELWALL_CHAIN_CUSTOM, 
		&chain)) != 0) {
		yyerror("unable to add chain '%s': %s\n", $2, strerror(-ret));
		free($2); 
		YYERROR;
	}

	free($2); 
}                 
                  filter_chain_c endl 
                | EINGABE 
{
	if((ret = amwall_add_chain(table, "INPUT", AMSELWALL_CHAIN_BUILTIN,
		&chain)) != 0) {
		yyerror("unable to add chain 'INPUT': %s\n", strerror(-ret));
		YYERROR;
	}
} 
                  filter_chain_c endl
                | WEITERLEITUNG 
{ 
	if((ret = amwall_add_chain(table, "FORWARD", AMSELWALL_CHAIN_BUILTIN,
		&chain)) != 0) {
		yyerror("unable to add chain 'FORWARD': %s\n", strerror(-ret));
		YYERROR;
	}
} 
                  filter_chain_c endl
                | AUSGABE
{ 
	if((ret = amwall_add_chain(table, "OUTPUT", AMSELWALL_CHAIN_BUILTIN,
		&chain)) != 0) {
		yyerror("unable to add chain 'OUTPUT': %s\n", strerror(-ret));
		YYERROR;
	}
} 
                  filter_chain_c endl
                | error endl
{ 
	errors++; 
}
		;

filter_chain_c  : optnl '{' optnl filter_chain '}'
                ;

filter_chain    : filter_chain filter_chain_l
                | filter_chain nl
                | filter_chain_l
                ;

filter_chain_l  : /* nothing */
                | AKZEPTIERE
{
	if((ret = amwall_add_rule(chain, "ACCEPT", AMSELWALL_RULE_ADD, 
		&rule)) != 0) {
		yyerror("unable to add rule: %s\n", strerror(-ret));
		YYERROR;
	}
}
                  filter_target_c punkt
                | VERWERFE
{
	if((ret = amwall_add_rule(chain, "DROP", AMSELWALL_RULE_ADD, 
		&rule)) != 0) {
		yyerror("unable to add rule: %s\n", strerror(-ret));
		YYERROR;
	}
}
                  filter_target_c punkt
                | PROTOKOLLIERE
{
	if((ret = amwall_add_rule(chain, "LOG", AMSELWALL_RULE_ADD, 
		&rule)) != 0) {
		yyerror("unable to add rule: %s\n", strerror(-ret));
		YYERROR;
	}
}
                  filter_target_c punkt
                | VERSUCHE string FUER
{
	if((ret = amwall_add_rule(chain, $2, AMSELWALL_RULE_ADD, 
		&rule)) != 0) {
		yyerror("unable to add rule: %s\n", strerror(-ret));
		free($2);
		YYERROR;
	}

	free($2);
}
                  filter_target_c punkt
                | WEISE
{
	if((ret = amwall_add_rule(chain, "REJECT", AMSELWALL_RULE_ADD, 
		&rule)) != 0) {
		yyerror("unable to add rule: %s\n", strerror(-ret));
		YYERROR;
	}
}
                  filter_target_c AB punkt
                | SETZE DIE KETTE string EIN punkt
{
	if((ret = amwall_add_rule(chain, $4, AMSELWALL_RULE_ADD, NULL)) != 0) {
		yyerror("unable to add rule: %s\n", strerror(-ret));
		free($4);
		YYERROR;
	}

	free($4);
}
                | VERWENDE DEN GRUNDSATZ policy punkt
{
	if((ret = amwall_add_rule(chain, $4, AMSELWALL_RULE_POLICY, NULL)) != 0) {
		yyerror("unable to add rule: %s\n", strerror(-ret));
		free($4);
		YYERROR;
	}

	free($4);
}
                | error punkt
{ 
	errors++; 
}
                ;

filter_target_c : condition VON source ZU dest options match fragment
                ;
 
policy          : AKZEPTIEREN
{
	$$ = strdup("ACCEPT");
}
                | VERWERFEN
{
	$$ = strdup("DROP");
}
                | string
{
	/* possibly not supported, but try... */
	$$ = $1;
}
                ;

condition       : ALLES
                | negation_c das PROTOKOLL string
{
	if((ret = amwall_add_entry(rule, $4, $1, AMSELWALL_ENTRY_CONDITION_PROTO, 
		NULL)) != 0) {
		yyerror("unable to add entry: %s\n", strerror(-ret));
		free($4);
		YYERROR;
	}

	free($4);
}
                ;

source          : source_t
                | source UND source_t
                ;

source_t        : IRGENDWOHER
                | negation source_l
{
	if((ret = amwall_add_entry(rule, $2, $1, entry_type, NULL)) != 0) {
		yyerror("unable to add entry: %s\n", strerror(-ret));
		free($2);
		YYERROR;
	}

	free($2);
}
                ;

source_l        : der QUELLE string
{
	entry_type = AMSELWALL_ENTRY_FROM_SOURCE;
	$$ = $3;
}
                | der SCHNITTSTELLE string
{
	entry_type = AMSELWALL_ENTRY_FROM_IFACE;
	$$ = $3;
}
                | dem PORT string
{
	entry_type = AMSELWALL_ENTRY_FROM_PORT;
	$$ = $3;
}
                ;

dest            : dest_t
                | dest UND dest_t
                ;

dest_t          : IRGENDEINEM ZIEL
                | negation dest_l
{
	if((ret = amwall_add_entry(rule, $2, $1, entry_type, 
		NULL)) != 0) {
		yyerror("unable to add entry: %s\n", strerror(-ret));
		free($2);
		YYERROR;
	}

	free($2);
}
                ;

dest_l          : dem ZIEL string
{
	entry_type = AMSELWALL_ENTRY_TO_DEST;
	$$ = $3;
}
                | der SCHNITTSTELLE string
{
	entry_type = AMSELWALL_ENTRY_TO_IFACE;
	$$ = $3;
}
                | dem PORT string
{
	entry_type = AMSELWALL_ENTRY_TO_PORT;
	$$ = $3;
}
                ;

match           : /* nothing */
                | IM FALLE VON string '(' optstring ')'
{
	if((ret = amwall_add_entry(rule, $4, 0, AMSELWALL_ENTRY_MATCH, 
		NULL)) != 0) {
		yyerror("unable to add entry: %s\n", strerror(-ret));
		free($4);
		if($6 != NULL)
			free($6);
		YYERROR;
	}
	
	if($6 != NULL) {
		if((ret = amwall_add_entry(rule, $6, 0, AMSELWALL_ENTRY_MISC, 
			NULL)) != 0) {
			yyerror("unable to add entry: %s\n", strerror(-ret));
			free($4);
			free($6);
			YYERROR;
		}

		free($6);
	}
	
	free($4);
}
                ;

fragment        : /* nothing */
                | ',' ABER NUR FRAGMENTE
{
	if((ret = amwall_add_entry(rule, "fragments", 0, 
		AMSELWALL_ENTRY_FRAGMENT, NULL)) != 0) {
		yyerror("unable to add entry: %s\n", strerror(-ret));
		YYERROR;
	}
}
                | ',' ABER KEINE FRAGMENTE
{
	if((ret = amwall_add_entry(rule, "coherent", 1, 
		AMSELWALL_ENTRY_FRAGMENT, NULL)) != 0) {
		yyerror("unable to add entry: %s\n", strerror(-ret));
		YYERROR;
	}
}
                ;

options         : /* nothing */
                | MIT DEN OPTIONEN string 
{
	if((ret = amwall_add_rule_options(rule, $4)) != 0) {
		yyerror("unable to add rule options: %s\n", strerror(-ret));
		free($4);
		YYERROR;
	}

	free($4);
}
                ;

negation        : /* nothing */
{
	$$ = 0;
}
                | BIS AUF
{
	$$ = 1;
}
                ;

negation_c      : /* nothing */
{
	$$ = 0;
}
                | ALLES BIS AUF
{
	$$ = 1;
}
                ;

der             : /* nothing */
                | DER
                ;

das             : /* nothing */
                | DAS
                ;

dem             : /* nothing */
                | DEM
                ;

punkt           : '.'
                | '.' '\n'
                ;

endl            : ';'
                | ';' '\n'
                ;

optnl		: /* nothing */
                | '\n' optnl
		;

optstring       : /* nothing */
{
	$$ = NULL;
}
                | string
{
	$$ = $1;
}
                ;

nl		: '\n' optnl		/* one newline or more */
		;
%%

/* this has to be sorted always */
const struct keywords keywords[] = {
        { "ab", AB },
        { "aber", ABER },
        { "akzeptiere", AKZEPTIERE },
        { "akzeptieren", AKZEPTIEREN },
        { "alle", ALLE },
        { "alles", ALLES },
        { "auf", AUF },
        { "ausgabe", AUSGABE },
        { "bis", BIS },
        { "das", DAS },
        { "dem", DEM },
        { "den", DEN },
        { "der", DER },
        { "die", DIE },
        { "ein", EIN },
        { "eingabe", EINGABE },
        { "ethernet", ETHERNET },
        { "falle", FALLE },
        { "filter", FILTER },
        { "fragmente", FRAGMENTE },
        { "fuer", FUER },
        { "grundsatz", GRUNDSATZ },
        { "im", IM },
        { "ip", IPV4 },
        { "ip6", IPV6 },
        { "irgendeinem", IRGENDEINEM },
        { "irgendwoher", IRGENDWOHER },
        { "keine", KEINE },
        { "kette", KETTE },
        { "mangle", MANGLE },
        { "mit", MIT },
        { "nat", NAT },
        { "nur", NUR },
        { "optionen", OPTIONEN },
        { "port", PORT },
        { "protokoll", PROTOKOLL },
        { "protokolliere", PROTOKOLLIERE },
        { "quelle", QUELLE },
        { "schnittstelle", SCHNITTSTELLE },
        { "setze", SETZE },
        { "und", UND },
        { "versuche", VERSUCHE },
        { "verwende", VERWENDE },
        { "verwerfe", VERWERFE },
        { "verwerfen", VERWERFEN },
        { "von", VON },
        { "weise", WEISE },
        { "weiterleitung", WEITERLEITUNG },
        { "wenn", WENN },
        { "ziel", ZIEL },
        { "zu", ZU },
};

const size_t keywords_size = sizeof(keywords)/sizeof(keywords[0]);
