/* -*- knf -*- */
/* $AMSEL: grammar_en.y,v 1.5 2004/05/23 15:03:38 reyk Exp $ */

/*
 * Copyright (c) 2004 .vantronix | secure systems GmbH
 * Copyright (c) 2004 SSV Embedded Systems GmbH
 *
 * All rights reserved.
 *
 * Written by Reyk Floeter <reyk@vantronix.net>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  *
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

%{
#include "main.h"
#include "tables.h"
#include "ipvs.h"
#include "parser.h"
#include "grammar.h"

FILE *fin = NULL;
int lineno = 1;
int errors = 0;
char *infile;

static int ret, option = 0;
static enum amselwall_table_type table_type;
static enum amselwall_chain_type chain_type;
static enum amselwall_entry_type entry_type;
static struct amselwall_table *table = NULL;
static struct amselwall_chain *chain = NULL;
static struct amselwall_rule *rule = NULL;
static struct amselwall_ipvs_real *real = NULL;
static struct amselwall_ipvs_server server;
static struct amselwall_ipvs_virtual *virtual = NULL;

#define AMWALL_ADD_RULE(_x) \
if((ret = amwall_add_rule(chain, _x, AMSELWALL_RULE_ADD, \
  &rule)) != 0) { \
    yyerror("Unable to add rule '%s': %s\n", _x, strerror(-ret)); \
    YYERROR; \
}

#define AMWALL_ADD_CHAIN(_x) \
chain_type = AMSELWALL_CHAIN_##_x; \
if((ret = amwall_add_chain(table, #_x, chain_type, \
  &chain)) != 0) { \
    yyerror("Unable to add chain '%s': %s\n", _x, strerror(-ret)); \
    YYERROR; \
}

#define AMWALL_ADD_TABLE(_x) \
if((ret = amwall_add_table(_x, table_type, &table)) != 0) { \
  yyerror("Unable to add table '%s': %s\n", _x, strerror(-ret)); \
  YYERROR; \
}

#define AMWALL_ADD_REAL(_x) \
if((ret = amwall_ipvs_add_real(table, _x, &real)) != 0) { \
  yyerror("Unable to add ipvs service '%s': %s\n", _x, strerror(-ret)); \
  YYERROR; \
}

#define AMWALL_ADD_SERVER(_x) \
if((ret = amwall_ipvs_add_server(real, _x, &server)) != 0) { \
  yyerror("Unable to add ipvs real server '%s': %s\n", _x, strerror(-ret)); \
  YYERROR; \
}

#define AMWALL_ADD_VIRTUAL(_x) \
if((ret = amwall_ipvs_add_virtual(table, _x, &virtual)) != 0) { \
  yyerror("Unable to add ipvs virtual server for '%s': %s\n", _x, strerror(-ret)); \
  YYERROR; \
}

%}

/* Generic tokens */
%token AND SET CHAIN BRIDGE IPV4 IPV6 FILTER NAT MANGLE BROUTE BROUTING
%token INPUT OUTPUT FORWARD PREROUTING POSTROUTING INSERT POLICY CONTINUE
%token ACCEPT DROP REJECT EXCEPT FRAGMENTS COHERENT TRY QUEUE RETURN LOGICAL
%token LOG ULOG REDIRECT SNAT DNAT MASQUERADE TCPMSS MARK DSCP ECN TOS
%token PROTO ANY FROM TO PORT SOURCE DEST IFACE IF WITH OPTIONS MIRROR

/* Token's introduced by the IPVS rules */
%token USE PERSISTENT SERVICE FOR SCHEDULER SERVER WEIGHT UPPER 
%token LOWER THRESHOLD ROUTE TUNNEL ON TIMEOUT NETMASK IPVS

%token <v.string> STRING
%type <v.string> string source_l dest_l optstring policy
%type <v.number> negation logical number
%%

/* 
 * Grammar
 */

grammar		: /* nothing */
                | grammar nl
                | grammar endl
                | grammar varset endl
		| grammar ipv4 endl
		| grammar ipv6 endl
		| grammar bridge endl
                | grammar ipvs endl
		| grammar error endl 
{ 
	errors++; 
}
		;

ipv4            : IPV4
{
	table_type = AMSELWALL_TABLE_IPV4;
}
                  ipv4_tables
                ;

ipv4_tables     : filter
                | nat
                | mangle
                ;

ipv4_ctargets   : accept target_rule
                | drop target_rule
                | queue target_rule
                | Return target_rule
                | log target_rule
                | ulog target_rule
                | mirror target_rule
                | reject target_rule
                | tcpmss target_rule
                | try target_rule
                | insert_chain
                | set_policy
                ;

ipv6            : IPV6
{
	table_type = AMSELWALL_TABLE_IPV6;
}
                  ipv6_tables
                ;

ipv6_tables     : filter6
                | mangle6
                ;

ipv6_ctargets   : accept target_rule
                | drop target_rule
                | queue target_rule
                | Return target_rule
                | log target_rule
                | reject target_rule
                | try target_rule
                | insert_chain
                | set_policy
                ;

bridge          : BRIDGE
{
	table_type = AMSELWALL_TABLE_BRIDGE;
}
                  bridge_tables
                ;

bridge_tables   : br_filter
                | br_nat
                | br_broute
                ;

bridge_ctargets : accept target_rule
                | drop target_rule
                | Continue target_rule
                | Return target_rule
		| mark target_rule
                | try target_rule
                | insert_chain
                | set_policy
                ;

/*
 * Misc types
 */

string		: string STRING				
{
	if (asprintf(&$$, "%s %s", $1, $2) == -1) {
		free($1);
		free($2);
		fatal("string: asprintf");
	}
	free($1);
	free($2);
}
		| STRING
		;

varset		: STRING '=' string		
{
	if (amwall_symset($1, $3, 0) == -1) {
		free($1);
		free($3);
		fatal("cannot store variable");
	}
	
	free($1);
	free($3);
}
		;

/*
 * IPv4 table "filter"
 */

filter          : FILTER optnl '{' optnl
{
	AMWALL_ADD_TABLE("filter");
}
                  filter_opts '}'
{
	table = NULL;
}
		;

filter_opts     : filter_opts filter_opts_l
                | filter_opts nl
                | filter_opts_l
                ;

filter_opts_l   : /* nothing */
                | chain optnl '{' optnl filter_chain '}' endl 
                | input optnl '{' optnl filter_chain '}' endl
                | forward optnl '{' optnl filter_chain '}' endl
                | output optnl '{' optnl filter_chain '}' endl
                | error endl
{ 
	errors++; 
}
		;

filter_chain    : filter_chain filter_chain_l
                | filter_chain nl
                | filter_chain_l
                ;

filter_chain_l  : /* nothing */
                | ipv4_ctargets endl
                | error endl
{ 
	errors++; 
}
                ;

/*
 * IPv4 table "nat"
 */

nat             : NAT optnl '{' optnl
{
	AMWALL_ADD_TABLE("nat");
}
                  nat_opts '}'
{
	table = NULL;
}
		;

nat_opts        : nat_opts nat_opts_l
                | nat_opts nl
                | nat_opts_l
                ;

nat_opts_l      : /* nothing */
                | chain optnl '{' optnl nat_chain '}' endl 
                | prerouting optnl '{' optnl nat_chain '}' endl
                | postrouting optnl '{' optnl nat_chain '}' endl
                | output optnl '{' optnl nat_chain '}' endl
                | error endl
{ 
	errors++; 
}
		;

nat_chain       : nat_chain nat_chain_l
                | nat_chain nl
                | nat_chain_l
                ;

nat_chain_l     : /* nothing */
                | ipv4_ctargets endl
                | dnat target_rule endl
                | snat target_rule endl
                | masquerade target_rule endl
                | redirect target_rule endl
                | error endl
{ 
	errors++; 
}
                ;

/*
 * IPv4 table "mangle"
 */

mangle          : MANGLE optnl '{' optnl
{
	AMWALL_ADD_TABLE("mangle");
}
                  mangle_opts '}'
{
	table = NULL;
}
		;

mangle_opts     : mangle_opts mangle_opts_l
                | mangle_opts nl
                | mangle_opts_l
                ;

mangle_opts_l   : /* nothing */
                | chain optnl '{' optnl mangle_chain '}' endl 
                | input '{' optnl mangle_chain '}' endl
                | output '{' optnl mangle_chain '}' endl
                | forward '{' optnl mangle_chain '}' endl
                | prerouting '{' optnl mangle_chain '}' endl
                | postrouting '{' optnl mangle_chain '}' endl
                | error endl
{ 
	errors++; 
}
		;

mangle_chain    : mangle_chain mangle_chain_l
                | mangle_chain nl
                | mangle_chain_l
                ;

mangle_chain_l  : /* nothing */
                | ipv4_ctargets endl
                | mark target_rule endl
                | dscp target_rule endl
                | ecn target_rule endl
                | tos target_rule endl
                | error endl
{ 
	errors++; 
}
                ;

/*
 * IPv6 table "filter"
 */


filter6         : FILTER optnl '{' optnl
{
	AMWALL_ADD_TABLE("filter");
}
                  filter6_opts '}'
{
	table = NULL;
}
		;

filter6_opts    : filter6_opts filter6_opts_l
                | filter6_opts nl
                | filter6_opts_l
                ;

filter6_opts_l  : /* nothing */
                | chain optnl '{' optnl filter6_chain '}' endl 
                | input optnl '{' optnl filter6_chain '}' endl
                | forward optnl '{' optnl filter6_chain '}' endl
                | output optnl '{' optnl filter6_chain '}' endl
                | error endl
{ 
	errors++; 
}
		;

filter6_chain   : filter6_chain filter6_chain_l
                | filter6_chain nl
                | filter6_chain_l
                ;

filter6_chain_l : /* nothing */
                | ipv6_ctargets endl
                | error endl
{ 
	errors++; 
}
                ;

/*
 * IPv6 table "mangle"
 */

mangle6         : MANGLE optnl '{' optnl
{
	AMWALL_ADD_TABLE("mangle");
}
                  mangle6_opts '}'
{
	table = NULL;
}
		;

mangle6_opts    : mangle6_opts mangle6_opts_l
                | mangle6_opts nl
                | mangle6_opts_l
                ;

mangle6_opts_l  : /* nothing */
                | chain optnl '{' optnl mangle6_chain '}' endl 
                | output '{' optnl mangle6_chain '}' endl
                | input '{' optnl mangle6_chain '}' endl
                | forward '{' optnl mangle6_chain '}' endl
                | prerouting '{' optnl mangle6_chain '}' endl
                | postrouting '{' optnl mangle6_chain '}' endl
                | error endl
{ 
	errors++; 
}
		;

mangle6_chain   : mangle6_chain mangle6_chain_l
                | mangle6_chain nl
                | mangle6_chain_l
                ;

mangle6_chain_l : /* nothing */
                | ipv6_ctargets endl
                | mark target_rule endl
                | error endl
{ 
	errors++; 
}
                ;

/*
 * Ethernet bridging table "filter"
 */

br_filter       : FILTER optnl '{' optnl
{
	AMWALL_ADD_TABLE("filter");
}
                  br_filter_opts '}'
{
	table = NULL;
}
		;

br_filter_opts  : br_filter_opts br_filter_opts_l
                | br_filter_opts nl
                | br_filter_opts_l
                ;

br_filter_opts_l: /* nothing */
                | chain optnl '{' optnl br_filter_chain '}' endl 
                | input optnl '{' optnl br_filter_chain '}' endl
                | forward optnl '{' optnl br_filter_chain '}' endl
                | output optnl '{' optnl br_filter_chain '}' endl
                | error endl
{ 
	errors++; 
}
		;

br_filter_chain : br_filter_chain br_filter_chainl
                | br_filter_chain nl
                | br_filter_chainl
                ;

br_filter_chainl: /* nothing */
                | bridge_ctargets endl
                | error endl
{ 
	errors++; 
}
                ;

/*
 * Ethernet bridging table "nat"
 */

br_nat          : NAT optnl '{' optnl
{
	AMWALL_ADD_TABLE("nat");
}
                  br_nat_opts '}'
{
	table = NULL;
}
		;

br_nat_opts     : br_nat_opts br_nat_opts_l
                | br_nat_opts nl
                | br_nat_opts_l
                ;

br_nat_opts_l: /* nothing */
                | chain optnl '{' optnl br_nat_chain '}' endl 
                | output optnl '{' optnl br_nat_chain '}' endl
                | prerouting optnl '{' optnl br_nat_chain '}' endl
                | postrouting optnl '{' optnl br_nat_chain '}' endl
                | error endl
{ 
	errors++; 
}
		;

br_nat_chain    : br_nat_chain br_nat_chain_l
                | br_nat_chain nl
                | br_nat_chain_l
                ;

br_nat_chain_l  : /* nothing */
                | bridge_ctargets endl
		| snat target_rule endl
		| dnat target_rule endl
		| redirect target_rule endl
                | error endl
{ 
	errors++; 
}
                ;

/*
 * Ethernet bridging table "broute"
 */

br_broute       : BROUTE optnl '{' optnl
{
	AMWALL_ADD_TABLE("broute");
}
                  br_broute_opts '}'
{
	table = NULL;
}
		;

br_broute_opts  : br_broute_opts br_broute_opts_l
                | br_broute_opts nl
                | br_broute_opts_l
                ;

br_broute_opts_l: /* nothing */
                | chain optnl '{' optnl br_broute_chain '}' endl 
                | brouting optnl '{' optnl br_broute_chain '}' endl
                | error endl
{ 
	errors++; 
}
		;

br_broute_chain : br_broute_chain br_broute_chainl
                | br_broute_chain nl
                | br_broute_chainl
                ;

br_broute_chainl: /* nothing */
                | bridge_ctargets endl
		| dnat target_rule endl
		| redirect target_rule endl
                | error endl
{ 
	errors++; 
}
                ;

/*
 * Common chains
 */

input           : INPUT
{
	AMWALL_ADD_CHAIN(INPUT);
} 
                ;

forward         : FORWARD 
{ 
	AMWALL_ADD_CHAIN(FORWARD);
} 
                ;

output          : OUTPUT
{ 
	AMWALL_ADD_CHAIN(OUTPUT);
} 
                ;

prerouting      : PREROUTING
{ 
	AMWALL_ADD_CHAIN(PREROUTING);
} 
                ;

postrouting     : POSTROUTING
{ 
	AMWALL_ADD_CHAIN(POSTROUTING);
} 
                ;

brouting        : BROUTING
{ 
	AMWALL_ADD_CHAIN(BROUTING);
} 
                ;

chain           : CHAIN string 
{ 
	chain_type = AMSELWALL_CHAIN_CUSTOM;
	if((ret = amwall_add_chain(table, $2, chain_type, &chain)) != 0) {
		yyerror("Unable to add chain '%s': %s", $2, strerror(-ret));
		free($2); 
		YYERROR;
	}

	free($2); 
}                 
                ;

/*
 * Common targets
 */

accept          : ACCEPT 
{
	AMWALL_ADD_RULE("ACCEPT");
}
                ;

Continue        : CONTINUE
{
	AMWALL_ADD_RULE("CONTINUE");
}

dnat            : DNAT
{
	switch(chain_type) {
	case AMSELWALL_CHAIN_CUSTOM:
	case AMSELWALL_CHAIN_PREROUTING:
	case AMSELWALL_CHAIN_OUTPUT:
	case AMSELWALL_CHAIN_BROUTING:
		break;
	default:
		yyerror("'dnat' not support in this chain");
		YYERROR;
	}

	AMWALL_ADD_RULE("DNAT");
}

drop            : DROP
{
	AMWALL_ADD_RULE("DROP");
}
                ;

dscp            : DSCP
{
	AMWALL_ADD_RULE("DSCP");
}
                ;

ecn             : ECN
{
	AMWALL_ADD_RULE("ECN");
}
                ;

queue           : QUEUE
{
	AMWALL_ADD_RULE("QUEUE");
}
                ;

Return          : RETURN
{
	AMWALL_ADD_RULE("RETURN");
}
                ;

log             : LOG
{
	AMWALL_ADD_RULE("LOG");
}
                ;

mark            : MARK
{
	AMWALL_ADD_RULE("MARK");
}
                ;

masquerade      : MASQUERADE
{
	switch(chain_type) {
	case AMSELWALL_CHAIN_CUSTOM:
	case AMSELWALL_CHAIN_POSTROUTING:
		break;
	default:
		yyerror("'masquerade' not support in this chain");
		YYERROR;
	}

	AMWALL_ADD_RULE("MASQUERADE");
}
                ;

mirror          : MIRROR
{
	switch(chain_type) {
	case AMSELWALL_CHAIN_CUSTOM:
	case AMSELWALL_CHAIN_INPUT:
	case AMSELWALL_CHAIN_FORWARD:
	case AMSELWALL_CHAIN_PREROUTING:
		break;
	default:
		yyerror("'mirror' not support in this chain");
		YYERROR;
	}

	AMWALL_ADD_RULE("MASQUERADE");
}
                ;

ulog            : ULOG
{
	AMWALL_ADD_RULE("ULOG");
}
                ;

redirect        : REDIRECT
{
	switch(chain_type) {
	case AMSELWALL_CHAIN_CUSTOM:
	case AMSELWALL_CHAIN_BROUTING:
	case AMSELWALL_CHAIN_PREROUTING:
		break;
	case AMSELWALL_CHAIN_OUTPUT:
		if(table_type == AMSELWALL_TABLE_BRIDGE) {
			yyerror("'redirect' not support in this bridge chain");
			YYERROR;
		}
		break;
	default:
		yyerror("'redirect' not support in this chain");
		YYERROR;
	}

	AMWALL_ADD_RULE("REDIRECT");
}
                ;

reject          : REJECT
{
	switch(chain_type) {
	case AMSELWALL_CHAIN_CUSTOM:
	case AMSELWALL_CHAIN_INPUT:
	case AMSELWALL_CHAIN_FORWARD:
	case AMSELWALL_CHAIN_OUTPUT:
		break;
	default:
		yyerror("'reject' not support in this chain");
		YYERROR;
	}

	AMWALL_ADD_RULE("REJECT");
}
                ;

snat            : SNAT
{
	switch(chain_type) {
	case AMSELWALL_CHAIN_CUSTOM:
	case AMSELWALL_CHAIN_POSTROUTING:
		break;
	default:
		yyerror("'snat' not support in this chain");
		YYERROR;
	}

	AMWALL_ADD_RULE("SNAT");
}
                ;

tcpmss          : TCPMSS
{
	AMWALL_ADD_RULE("TCPMSS");
}
                ;

tos             : TOS
{
	AMWALL_ADD_RULE("TOS");
}
                ;

try             : TRY string
{
	if((ret = amwall_add_rule(chain, $2, AMSELWALL_RULE_ADD, 
		&rule)) != 0) {
		yyerror("Unable to add rule: %s", strerror(-ret));
		free($2);
		YYERROR;
	}

	free($2);
}
                ;

insert_chain    : INSERT CHAIN string
{
	if((ret = amwall_add_rule(chain, $3, AMSELWALL_RULE_ADD, NULL)) != 0) {
		yyerror("Unable to add rule: %s", strerror(-ret));
		free($3);
		YYERROR;
	}

	free($3);
}
                ;

set_policy      : SET POLICY policy
{
	if((ret = amwall_add_rule(chain, $3, AMSELWALL_RULE_POLICY, NULL)) != 0) {
		yyerror("Unable to add rule: %s", strerror(-ret));
		free($3);
		YYERROR;
	}

	free($3);
}
                ;

/*
 * Common entries
 */

target_rule     : condition FROM source TO dest match options fragment
                ;

policy          : ACCEPT
{
	$$ = strdup("ACCEPT");
}
                | DROP
{
	$$ = strdup("DROP");
}
                | RETURN
{
	/* possibly only works with ebtables */
	$$ = strdup("RETURN");
}
                | string
{
	if(table_type == AMSELWALL_TABLE_BRIDGE) {
		yyerror("Policy '%s' is not supported in bridging tables!", $1);
		free($1);
		YYERROR;
	}

	/* possibly not supported, but try... */
	$$ = $1;
}
                ;

condition       : ANY
                | negation PROTO string
{
	if((ret = amwall_add_entry(rule, $3, $1, AMSELWALL_ENTRY_CONDITION_PROTO, 
		NULL)) != 0) {
		yyerror("Unable to add entry: %s", strerror(-ret));
		free($3);
		YYERROR;
	}

	free($3);
}
                ;

fragment        : /* nothing */
                | FRAGMENTS
{
	if((ret = amwall_add_entry(rule, "fragments", 0, 
		AMSELWALL_ENTRY_FRAGMENT, NULL)) != 0) {
		yyerror("Unable to add entry: %s", strerror(-ret));
		YYERROR;
	}
}
                | COHERENT
{
	if((ret = amwall_add_entry(rule, "coherent", 1, 
		AMSELWALL_ENTRY_FRAGMENT, NULL)) != 0) {
		yyerror("Unable to add entry: %s", strerror(-ret));
		YYERROR;
	}
}
                ;

source          : source_t
                | source AND source_t
                ;

source_t        : ANY
                | negation source_l
{
	if((ret = amwall_add_entry(rule, $2, $1, entry_type, NULL)) != 0) {
		yyerror("Unable to add entry: %s", strerror(-ret));
		free($2);
		YYERROR;
	}

	free($2);
}
                ;

source_l        : SOURCE string
{
	entry_type = AMSELWALL_ENTRY_FROM_SOURCE;
	$$ = $2;
}
                | logical IFACE string
{
	if(chain_type == POSTROUTING) {
		yyerror("Entry '...from iface %s...' " 
		    "is invalid in POSTROUTING chains!", $3);
		free($3);
		YYERROR;
	}

	if($1)
		entry_type = AMSELWALL_ENTRY_FROM_LOGICAL_IFACE;
	else
		entry_type = AMSELWALL_ENTRY_FROM_IFACE;
	$$ = $3;
}
                | PORT string
{
	if(table_type == AMSELWALL_TABLE_BRIDGE) {
		yyerror("Entry '..from port %s...' "
		    "is invalid bridge tables", $2);
		free($2);
		YYERROR;
	}

	entry_type = AMSELWALL_ENTRY_FROM_PORT;
	$$ = $2;
}
                ;

dest            : dest_t
                | dest AND dest_t
                ;

dest_t          : ANY
                | negation dest_l
{
	if((ret = amwall_add_entry(rule, $2, $1, entry_type, 
		NULL)) != 0) {
		yyerror("Unable to add entry: %s", strerror(-ret));
		free($2);
		YYERROR;
	}

	free($2);
}
                | AND dest
                ;

dest_l          : DEST string
{
	entry_type = AMSELWALL_ENTRY_TO_DEST;
	$$ = $2;
}
                | logical IFACE string
{
	if(chain_type == PREROUTING) {
		yyerror("Entry '...to iface %s...' " 
		    "is invalid in PREROUTING chains!", $3);
		free($3);
		YYERROR;
	}

	if($1)
		entry_type = AMSELWALL_ENTRY_TO_LOGICAL_IFACE;
	else
		entry_type = AMSELWALL_ENTRY_TO_IFACE;
	$$ = $3;
}
                | PORT string
{
	if(table_type == AMSELWALL_TABLE_BRIDGE) {
		yyerror("Entry '..to port %s...' "
		    "is invalid bridge tables", $2);
		free($2);
		YYERROR;
	}

	entry_type = AMSELWALL_ENTRY_TO_PORT;
	$$ = $2;
}
                ;

logical         : /* nothing */
{
	$$ = 0;
}
                | BRIDGE
{
	if(table_type != AMSELWALL_TABLE_BRIDGE) {
		yyerror("Entry '..bridge iface...' "
		    "only allow in bridge tables");
		YYERROR;
	}

	$$ = 1;
}
                ;

match           : /* nothing */
                | IF string '{' optstring '}'
{
	if((ret = amwall_add_entry(rule, $2, 0, AMSELWALL_ENTRY_MATCH, 
		NULL)) != 0) {
		yyerror("Unable to add entry: %s", strerror(-ret));
		free($2);
		if($4 != NULL)
			free($4);
		YYERROR;
	}
	
	if($4 != NULL) {
		if((ret = amwall_add_entry(rule, $4, 0, AMSELWALL_ENTRY_MISC, 
			NULL)) != 0) {
			yyerror("Unable to add entry: %s", strerror(-ret));
			free($2);
			free($4);
			YYERROR;
		}
	}
	
	free($2);
	if($4 != NULL)
		free($4);
}
                ;

options         : /* nothing */
                | WITH OPTIONS '{' string '}'
{
	if((ret = amwall_add_rule_options(rule, $4)) != 0) {
		yyerror("Unable to add rule 'with options': %s", strerror(-ret));
		free($4);
		YYERROR;
	}

	free($4);
}
                ;

negation        : /* nothing */
{
	$$ = 0;
}
                | ANY EXCEPT
{
	$$ = 1;
}
                ;

endl            : ';'
                | ';' '\n'
                ;

optnl		: /* nothing */
                | '\n' optnl
		;

optstring       : /* nothing */
{
	$$ = NULL;
}
                | string
{
	$$ = $1;
}
                ;

nl		: '\n' optnl		/* one newline or more */
		;


number		: STRING
{
	u_long ulval;

	if (atoul($1, &ulval) == -1) {
		yyerror("%s: not a number", $1);
		free($1);
		YYERROR;
	} else
		$$ = ulval;

	free($1);
}
		;

and             : AND
                | ','
                ;

/*
 * IPVS (Linux IP Virtual Server) is a bit different from the others...
 */

ipvs            : IPVS optnl '{' optnl
{
	table_type = AMSELWALL_TABLE_IPVS;
	AMWALL_ADD_TABLE("ipvs");
}
                  ipvs_opts '}'
{
	table = NULL;
}
		;

ipvs_opts       : ipvs_opts ipvs_opts_l
                | ipvs_opts nl
                | ipvs_opts_l
                ;

ipvs_opts_l     : /* nothing */
                | SERVICE string optnl '{' optnl
{
	AMWALL_ADD_REAL($2);
	free($2);
}
                  ipvs_service '}' endl
{
	real = NULL;
}
                | USE ipvs_persistent SERVICE string

{
	AMWALL_ADD_VIRTUAL($4);
	free($4);
}
                  FOR ipvs_match ipvs_vflags endl
{
	if(option)
		virtual->flags |= AMSELWALL_IPVS_VFLAG_PERSISTENT;
	option = 0;
}
                ;

ipvs_persistent : /* nothing */
                | PERSISTENT
{
	option = 1;
}
                ;

ipvs_vflags     : /* nothing */
                | WITH ipvs_vflags_t
                ; 

ipvs_vflags_t   : ipvs_vflags_l
                | ipvs_vflags_t and ipvs_vflags_l
                ;

ipvs_vflags_l   : SCHEDULER string
{
	virtual->flags |= AMSELWALL_IPVS_VFLAG_SCHEDULER;
	if((virtual->scheduler = strdup($2)) == NULL) {
		yyerror("Failed to add virtual server scheduler '%s': %s\n", 
		    $2, strerror(errno));
		free($2);
		YYERROR;
	}
	free($2);
}
                | TIMEOUT number
{
	if(!option) {
		yyerror("IPVS option 'timeout' is only valid for persistent schedulers\n");
		YYERROR;
	}
	virtual->flags |= AMSELWALL_IPVS_VFLAG_TIMEOUT;
	virtual->timeout = $2;
}
                | NETMASK string
{
	if(!option) {
		yyerror("IPVS option 'netmask' is only valid for persistent schedulers\n");
		free($2);
		YYERROR;
	}
	virtual->flags |= AMSELWALL_IPVS_VFLAG_NETMASK;
	if((virtual->netmask = strdup($2)) == NULL) {
		yyerror("Failed to add virtual server netmask '%s': %s\n", 
		    $2, strerror(errno));
		free($2);
		YYERROR;
	}
	free($2);
}
                ;

ipvs_match      : MARK number
{
	virtual->type = AMSELWALL_IPVS_VTYPE_MARK;
	virtual->mark = $2;
}
                | PROTO string ON string ipvs_match_port
{
	if(strcasecmp("tcp", $2) == 0)
		virtual->type = AMSELWALL_IPVS_VTYPE_TCP;
	else if(strcasecmp("udp", $2) == 0)
		virtual->type = AMSELWALL_IPVS_VTYPE_UDP;
	else {
		yyerror("IPVS not supported for protocol '%s'\n", $2);
		free($2);
		YYERROR;
	}

	if((virtual->address = strdup($4)) == NULL) {
		yyerror("Failed to add virtual server address '%s': %s\n", 
		    $4, strerror(errno));
		free($4);
		YYERROR;
	}

	free($4);
	free($2);
}
                ; 

ipvs_match_port : /* nothing */
                | AND PORT number
{
	virtual->flags |= AMSELWALL_IPVS_VFLAG_PORT;
	virtual->port = $3;
}
                ;

ipvs_service    : ipvs_service ipvs_service_l
                | ipvs_service nl
                | ipvs_service_l
                ;

ipvs_service_l  : /* nothing */
                | ipvs_method TO SERVER string 
{
	server.flags = 0;
}
                  ipvs_port ipvs_flags endl
{
	AMWALL_ADD_SERVER($4);
	free($4);
}
                ;

ipvs_method     : MASQUERADE
{
	server.type = AMSELWALL_IPVS_RTYPE_MASQUERADE;
}
                | TUNNEL
{
	server.type = AMSELWALL_IPVS_RTYPE_TUNNEL;
}
                | ROUTE
{
	server.type = AMSELWALL_IPVS_RTYPE_ROUTE;
}
                ;

ipvs_port       : /* nothing */
                | ON PORT number
{
	server.flags |= AMSELWALL_IPVS_SFLAG_PORT;
	server.port = $3;
}
                ;

ipvs_flags      : /* nothing */
                | WITH ipvs_flags_t
                ; 

ipvs_flags_t    : ipvs_flags_l
                | ipvs_flags_t and ipvs_flags_l
                ;

ipvs_flags_l    : WEIGHT number
{
	server.flags |= AMSELWALL_IPVS_SFLAG_WEIGHT;
	server.weight = $2;
}
                | LOWER THRESHOLD number
{
	server.flags |= AMSELWALL_IPVS_SFLAG_LTHR;
	server.lthreshold = $3;
}
                | UPPER THRESHOLD number
{
	server.flags |= AMSELWALL_IPVS_SFLAG_UTHR;
	server.uthreshold = $3;
}
                ;

%%

/* this has to be sorted always */
const struct keywords keywords[] = {
	{ "accept", ACCEPT },
        { "and", AND },
	{ "any", ANY },
	{ "bridge", BRIDGE },
	{ "broute", BROUTE },
	{ "brouting", BROUTING },
	{ "chain", CHAIN },
	{ "coherent", COHERENT },
	{ "continue", CONTINUE },
        { "dest", DEST },
	{ "dnat", DNAT },
	{ "drop", DROP },
	{ "dscp", DSCP },
	{ "ecn", ECN },
	{ "except", EXCEPT },
        { "filter", FILTER },
	{ "for", FOR },
        { "forward", FORWARD },
	{ "fragments", FRAGMENTS },
	{ "from", FROM },
	{ "if", IF },
	{ "iface", IFACE },
        { "input", INPUT },
        { "insert", INSERT },
        { "ip", IPV4 },
        { "ip6", IPV6 },
        { "ipvs", IPVS },
	{ "log", LOG },
	{ "lower", LOWER },
        { "mangle", MANGLE },
	{ "mark", MARK },
	{ "masquerade", MASQUERADE },
	{ "mirror", MIRROR },
        { "nat", NAT },
        { "netmask", NETMASK },
	{ "on", ON },
	{ "options", OPTIONS },
        { "output", OUTPUT },
	{ "persistent", PERSISTENT },
	{ "policy", POLICY },
	{ "port", PORT },
	{ "postrouting", POSTROUTING },
	{ "prerouting", PREROUTING },
	{ "proto", PROTO },
	{ "queue", QUEUE },
	{ "redirect", REDIRECT },
	{ "reject", REJECT },
 	{ "return", RETURN },
	{ "route", ROUTE },
	{ "scheduler", SCHEDULER },
	{ "server", SERVER },
	{ "service", SERVICE },
	{ "set", SET },
	{ "snat", SNAT },
	{ "source", SOURCE },
	{ "tcpmss", TCPMSS },
	{ "threshold", THRESHOLD },
	{ "timeout", TIMEOUT },
	{ "to", TO },
	{ "tos", TOS },
	{ "try", TRY },
	{ "tunnel", TUNNEL },
	{ "ulog", ULOG },
	{ "upper", UPPER },
	{ "use", USE },
	{ "weight", WEIGHT },
	{ "with", WITH },
};

const size_t keywords_size = sizeof(keywords)/sizeof(keywords[0]);
