/* -*- knf -*- */
/* $AMSEL: ipvs.c,v 1.2 2004/05/23 15:03:38 reyk Exp $ */

/*
 * Copyright (c) 2004 .vantronix | secure systems GmbH
 * Copyright (c) 2004 SSV Embedded Systems GmbH
 *
 * All rights reserved.
 *
 * Written by Reyk Floeter <reyk@vantronix.net>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  *
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "main.h"
#include "tables.h"
#include "ipvs.h"
#include "parser.h"
#include "version.h"

int
amwall_ipvs_add_server(struct amselwall_ipvs_real *real, 
    const char *address, struct amselwall_ipvs_server *options)
{
	struct amselwall_ipvs_server *server;

	if((server = (struct amselwall_ipvs_server*)calloc(1, 
		sizeof(struct amselwall_ipvs_server))) == NULL)
		return(-ENOMEM);

	if((server->address = strdup(address)) == NULL) {
		free(server);
		return(-ENOMEM);
	}

	server->flags = options->flags;
	server->type = options->type;

	if(server->flags & AMSELWALL_IPVS_SFLAG_PORT)
		server->port = options->port;

	if(server->flags & AMSELWALL_IPVS_SFLAG_WEIGHT)
		server->weight = options->weight;

	if(server->flags & AMSELWALL_IPVS_SFLAG_UTHR)
		server->uthreshold = options->uthreshold;

	if(server->flags & AMSELWALL_IPVS_SFLAG_LTHR)
		server->lthreshold = options->lthreshold;

	TAILQ_INSERT_TAIL(&real->h_head, server, entries);

	return(0);
}

int
amwall_ipvs_add_real(struct amselwall_table *table, 
    const char *name, struct amselwall_ipvs_real **real_ptr)
{
	struct amselwall_ipvs *ipvs;
	struct amselwall_ipvs_real *real;

	if(table->data == NULL)
		return(-EFAULT);

	ipvs = (struct amselwall_ipvs*)table->data;

	if((real = (struct amselwall_ipvs_real*)calloc(1, 
		sizeof(struct amselwall_ipvs_real))) == NULL)
		return(-ENOMEM);

	if((real->name = strdup(name)) == NULL) {
		free(real);
		return(-ENOMEM);
	}

	TAILQ_INIT(&real->h_head);
	TAILQ_INSERT_TAIL(&ipvs->h_rhead, real, entries);

	if(real_ptr != NULL)
		*real_ptr = real;

	return(0);
}

int
amwall_ipvs_add_virtual(struct amselwall_table *table, 
    const char *servers, struct amselwall_ipvs_virtual **virtual_ptr)
{
	struct amselwall_ipvs *ipvs;
	struct amselwall_ipvs_virtual *virtual;
	struct amselwall_ipvs_real *real;
	int found = 0;

	if(table->data == NULL)
		return(-EFAULT);

	ipvs = (struct amselwall_ipvs*)table->data;

	TAILQ_FOREACH(real, &ipvs->h_rhead, entries) {
		if(strcmp(real->name, servers) == 0) {
			found++;
			break;
		}
	}

	if(found != 1)
		return(-ENOENT);

	if((virtual = (struct amselwall_ipvs_virtual*)calloc(1, 
		sizeof(struct amselwall_ipvs_virtual))) == NULL)
		return(-ENOMEM);

	virtual->servers = real;

	TAILQ_INSERT_TAIL(&ipvs->h_vhead, virtual, entries);

	if(virtual_ptr != NULL)
		*virtual_ptr = virtual;

	return(0);
}

int 
amwall_ipvs_add_table(struct amselwall_table *table)
{
	struct amselwall_ipvs *ipvs;

	if((table->data = calloc(1, sizeof(struct amselwall_ipvs))) == NULL)
		return(-ENOMEM);

	ipvs = (struct amselwall_ipvs*)table->data;

	TAILQ_INIT(&ipvs->h_rhead);
	TAILQ_INIT(&ipvs->h_vhead);

	return(0);
}

void
amwall_ipvs_cleanup(struct amselwall_table *table)
{
	struct amselwall_ipvs *ipvs;
	struct amselwall_ipvs_real *real, *next_real;
	struct amselwall_ipvs_server *server, *next_server;
	struct amselwall_ipvs_virtual *virtual, *next_virtual;

	if(table->data == NULL)
		return;

	ipvs = (struct amselwall_ipvs*)table->data;

	for(real = TAILQ_FIRST(&ipvs->h_rhead); 
	    real != NULL; real = next_real) {
		next_real = TAILQ_NEXT(real, entries);
		TAILQ_REMOVE(&ipvs->h_rhead, real, entries);
		for(server = TAILQ_FIRST(&real->h_head); 
		    server != NULL; server = next_server) {
			next_server = TAILQ_NEXT(server, entries);
			TAILQ_REMOVE(&real->h_head, server, entries);
			free(server->address);
			free(server);
		}
		free(real->name);
		free(real);
	}

	for(virtual = TAILQ_FIRST(&ipvs->h_vhead); 
	    virtual != NULL; virtual = next_virtual) {
		next_virtual = TAILQ_NEXT(virtual, entries);
		TAILQ_REMOVE(&ipvs->h_vhead, virtual, entries);
		if(virtual->type > AMSELWALL_IPVS_VTYPE_MARK)
			free(virtual->address);
		if(virtual->flags & AMSELWALL_IPVS_VFLAG_SCHEDULER)
			free(virtual->scheduler);
		if(virtual->flags & AMSELWALL_IPVS_VFLAG_NETMASK)
			free(virtual->netmask);
		virtual->servers = NULL;
		free(virtual);
	}

	free(ipvs);
}

int
amwall_ipvs_generate(struct amselwall_table *table)
{
	struct amselwall_ipvs *ipvs;
	struct amselwall_ipvs_real *real;
	struct amselwall_ipvs_server *server;
	struct amselwall_ipvs_virtual *virtual;
	char *match = NULL;
	
	if(table->data == NULL)
		return(-EFAULT);

	ipvs = (struct amselwall_ipvs*)table->data;

	if(conf->flags & AMSELWALL_OPT_OPTIMIZE)
		printf("echo \"\n");

	if(!(conf->flags & AMSELWALL_OPT_KEEP) || conf->flags & AMSELWALL_OPT_FLUSH) 
		printf("%s -C\n", conf->flags & AMSELWALL_OPT_OPTIMIZE ? 
		    " " : AMSELWALL_PROG_IPVSADM);

	if(conf->flags & AMSELWALL_OPT_FLUSH) 
		goto done;

	TAILQ_FOREACH(virtual, &ipvs->h_vhead, entries) {
		if((real = virtual->servers) == NULL)
			continue;

		switch(virtual->type) {
		case AMSELWALL_IPVS_VTYPE_TCP:
			if(asprintf(&match, "-t %s:%u",
			       virtual->address,
			       virtual->flags & AMSELWALL_IPVS_VFLAG_PORT ?
			       virtual->port : 0) == -1)
				continue;
			break;
		case AMSELWALL_IPVS_VTYPE_UDP:
			if(asprintf(&match, "-u %s:%u",
			       virtual->address,
			       virtual->flags & AMSELWALL_IPVS_VFLAG_PORT ?
			       virtual->port : 0) == -1)
				continue;
			break;
		case AMSELWALL_IPVS_VTYPE_MARK:
			if(asprintf(&match, "-f %u", virtual->mark) == -1)
				continue;
			break;
		}

		printf("%s -%c %s ", 
		    conf->flags & AMSELWALL_OPT_OPTIMIZE ? " " : AMSELWALL_PROG_IPVSADM, 
		    conf->flags & AMSELWALL_OPT_EDIT ? 'E' : 'A', match);

		if(virtual->flags & AMSELWALL_IPVS_VFLAG_SCHEDULER)
			printf("-s %s ", virtual->scheduler);

		if(virtual->flags & AMSELWALL_IPVS_VFLAG_NETMASK)
			printf("-M %s ", virtual->netmask);

		if(virtual->flags & AMSELWALL_IPVS_VFLAG_PERSISTENT) {
			printf("-p");
			if(virtual->flags & AMSELWALL_IPVS_VFLAG_TIMEOUT)
				printf(" %u", virtual->timeout);
		}

		printf("\n");

		TAILQ_FOREACH(server, &real->h_head, entries) {
			printf("%s -%c %s -r %s:%u ",
			    conf->flags & AMSELWALL_OPT_OPTIMIZE ? " " : AMSELWALL_PROG_IPVSADM, 
			    conf->flags & AMSELWALL_OPT_EDIT ? 'e' : 'a', 
			    match, server->address,
			    server->flags & AMSELWALL_IPVS_SFLAG_PORT ? 
			    server->port : 0);
			if(server->flags & AMSELWALL_IPVS_SFLAG_WEIGHT)
				printf("-w %u ", server->weight);
			if(server->flags & AMSELWALL_IPVS_SFLAG_UTHR)
				printf("-x %u ", server->uthreshold);
			if(server->flags & AMSELWALL_IPVS_SFLAG_LTHR)
				printf("-y %u ", server->lthreshold);

			switch(server->type) {
			case AMSELWALL_IPVS_RTYPE_MASQUERADE:
				printf("-m\n");
				break;
			case AMSELWALL_IPVS_RTYPE_TUNNEL:
				printf("-i\n");
				break;
			case AMSELWALL_IPVS_RTYPE_ROUTE:
				printf("-g\n");
				break;
			}
		}
		free(match);
	}

 done:
	if(conf->flags & AMSELWALL_OPT_OPTIMIZE)
		printf("\" | %s -R\n", AMSELWALL_PROG_IPVSADM);

	return(0);
}
