/* -*- knf -*- */
/* $AMSEL: ipvs.h,v 1.1 2004/05/23 13:51:53 reyk Exp $ */

/*
 * Copyright (c) 2004 .vantronix | secure systems GmbH
 * Copyright (c) 2004 SSV Embedded Systems GmbH
 *
 * All rights reserved.
 *
 * Written by Reyk Floeter <reyk@vantronix.net>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  *
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#ifndef _AMSEL_WALL_IPVS_HEADER
#define _AMSEL_WALL_IPVS_HEADER

#define AMSELWALL_PROG_IPVSADM "/sbin/ipvsadm"

enum amselwall_ipvs_server_type {
	AMSELWALL_IPVS_RTYPE_MASQUERADE = 0,
	AMSELWALL_IPVS_RTYPE_TUNNEL,
	AMSELWALL_IPVS_RTYPE_ROUTE,
};

#define AMSELWALL_IPVS_SFLAG_PORT 0x01
#define AMSELWALL_IPVS_SFLAG_WEIGHT 0x02
#define AMSELWALL_IPVS_SFLAG_UTHR 0x04
#define AMSELWALL_IPVS_SFLAG_LTHR 0x08

struct amselwall_ipvs_server {
	char *address;
	enum amselwall_ipvs_server_type type;
	u_int8_t flags;
	u_int16_t port;
	u_int16_t weight;
	u_int16_t uthreshold;
	u_int16_t lthreshold;
	TAILQ_ENTRY(amselwall_ipvs_server) entries;
};

struct amselwall_ipvs_real {
	char *name;
	TAILQ_ENTRY(amselwall_ipvs_real) entries;
	TAILQ_HEAD(amselwall_ipvs_server_head, amselwall_ipvs_server) h_head;
};

enum amselwall_ipvs_virtual_type {
	AMSELWALL_IPVS_VTYPE_MARK = 0,
	AMSELWALL_IPVS_VTYPE_TCP,
	AMSELWALL_IPVS_VTYPE_UDP,
};

#define AMSELWALL_IPVS_VFLAG_PERSISTENT 0x01
#define AMSELWALL_IPVS_VFLAG_TIMEOUT 0x02
#define AMSELWALL_IPVS_VFLAG_SCHEDULER 0x04
#define AMSELWALL_IPVS_VFLAG_PORT 0x08
#define AMSELWALL_IPVS_VFLAG_NETMASK 0x10

struct amselwall_ipvs_virtual {
	enum amselwall_ipvs_virtual_type type;
	u_int8_t flags;
	char *address;
	char *scheduler;
	char *netmask;
	u_int32_t timeout;
	u_int16_t port;
	u_int16_t mark;
	struct amselwall_ipvs_real *servers;
	TAILQ_ENTRY(amselwall_ipvs_virtual) entries;
};

struct amselwall_ipvs {
	TAILQ_HEAD(amselwall_ipvs_real_head, amselwall_ipvs_real) h_rhead;
	TAILQ_HEAD(amselwall_ipvs_virtual_head, amselwall_ipvs_virtual) h_vhead;
};

extern int amwall_ipvs_add_table(struct amselwall_table *);
extern int amwall_ipvs_add_server(struct amselwall_ipvs_real *, 
    const char *, struct amselwall_ipvs_server *);
extern int amwall_ipvs_add_real(struct amselwall_table *, 
    const char *, struct amselwall_ipvs_real **);
extern int amwall_ipvs_add_virtual(struct amselwall_table *, 
    const char *, struct amselwall_ipvs_virtual **);
extern void amwall_ipvs_cleanup(struct amselwall_table *);
extern int amwall_ipvs_generate(struct amselwall_table *);

#endif /* _AMSEL_WALL_IPVS_HEADER */
