/* -*- knf -*- */
/* $AMSEL: tables.c,v 1.1 2004/05/23 15:03:38 reyk Exp $ */

/*
 * Copyright (c) 2004 .vantronix | secure systems GmbH
 * Copyright (c) 2004 SSV Embedded Systems GmbH
 *
 * All rights reserved.
 *
 * Written by Reyk Floeter <reyk@vantronix.net>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  *
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "main.h"
#include "tables.h"
#include "ipvs.h"
#include "parser.h"
#include "version.h"

/*
 * Add a new table
 */
int 
amwall_add_table(char *name, enum amselwall_table_type type,
    struct amselwall_table **table_ptr)
{
	int ret = 0;
	struct amselwall_table *table;

	if((table = (struct amselwall_table*)
	       calloc(1, sizeof(struct amselwall_table))) == NULL)
		return(-ENOMEM);

	table->type = type;

	if(type == AMSELWALL_TABLE_IPVS) {
		if((ret = amwall_ipvs_add_table(table)) != 0) {
			free(table);
			return(ret);
		}
	}

	if((table->name = strdup(name)) == NULL) {
		free(table);
		return(-ENOMEM);
	}

	TAILQ_INIT(&table->h_head);
	TAILQ_INSERT_TAIL(&conf->h_head, table, entries);

	if(table_ptr != NULL)
		*table_ptr = table;

	return(0);
}

/*
 * Add a new chain
 */
int 
amwall_add_chain(struct amselwall_table *table, char *name, 
    enum amselwall_chain_type type, struct amselwall_chain **chain_ptr)
{
	struct amselwall_chain *chain;

	if(table == NULL)
		return(-EFAULT);

	if((chain = (struct amselwall_chain*)
	       calloc(1, sizeof(struct amselwall_chain))) == NULL)
		return(-ENOMEM);

	if((chain->name = strdup(name)) == NULL) {
		free(chain);
		return(-ENOMEM);
	}

	chain->type = type;
	chain->table = table;

	TAILQ_INIT(&chain->h_head);
	TAILQ_INSERT_TAIL(&table->h_head, chain, entries);

	if(chain_ptr != NULL)
		*chain_ptr = chain;

	return(0);
}

/*
 * Add a new rule
 */
int 
amwall_add_rule(struct amselwall_chain *chain, char *target, 
    enum amselwall_table_type type, struct amselwall_rule **rule_ptr)
{
	struct amselwall_rule *rule;

	if(chain == NULL)
		return(-EFAULT);

	if((rule = (struct amselwall_rule*)
	       calloc(1, sizeof(struct amselwall_rule))) == NULL)
		return(-ENOMEM);

	if((rule->target = strdup(target)) == NULL) {
		free(rule);
		return(-ENOMEM);
	}

	rule->type = type;
	rule->chain = chain;
	rule->options = NULL;

	TAILQ_INIT(&rule->h_head);
	TAILQ_INSERT_TAIL(&chain->h_head, rule, entries);

	if(rule_ptr != NULL)
		*rule_ptr = rule;

	return(0);
}

/*
 * Add options for selected rule target
 */
int 
amwall_add_rule_options(struct amselwall_rule *rule, char *options)
{
	if((rule->options = strdup(options)) == NULL)
		return(-ENOMEM);

	return(0);
}

/*
 * Add a new entry
 */
int 
amwall_add_entry(struct amselwall_rule *rule, char *option, int negation,
    enum amselwall_entry_type type, struct amselwall_entry **entry_ptr)
{
	int ret = 0;
	struct amselwall_entry *entry;

	if(rule == NULL)
		return(-EFAULT);

	if(rule->type != AMSELWALL_RULE_ADD)
		return(-EINVAL);

	if((entry = (struct amselwall_entry*)
	       calloc(1, sizeof(struct amselwall_entry))) == NULL)
		return(-ENOMEM);

	entry->option = NULL;
	entry->negation = negation;

	switch(type) {
	case AMSELWALL_ENTRY_MISC:
		entry->option = strdup(option);
		break;
	case AMSELWALL_ENTRY_MATCH:
		ret = asprintf(&entry->option, "-m %s", option) ? 0 : -EINVAL;
		break;
	case AMSELWALL_ENTRY_FRAGMENT:
		ret = asprintf(&entry->option, "%s-f", 
		    entry->negation ? "! " : "") ? 0 : -EINVAL;
		break;
	case AMSELWALL_ENTRY_CONDITION_PROTO:
		ret = asprintf(&entry->option, "-p %s%s", 
		    entry->negation ? "! " : "", option) ? 0 : -EINVAL;
		break;
	case AMSELWALL_ENTRY_FROM_SOURCE:
		ret = asprintf(&entry->option, "-s %s%s",
		    entry->negation ? "! " : "", option) ? 0 : -EINVAL;
		break;
	case AMSELWALL_ENTRY_FROM_IFACE:
		ret = asprintf(&entry->option, "-i %s%s", 
		    entry->negation ? "! " : "", option) ? 0 : -EINVAL;
		break;
	case AMSELWALL_ENTRY_FROM_LOGICAL_IFACE:
		ret = asprintf(&entry->option, "--logical-in %s%s", 
		    entry->negation ? "! " : "", option) ? 0 : -EINVAL;
		break;
	case AMSELWALL_ENTRY_FROM_PORT:
		ret = asprintf(&entry->option, "--sport %s%s", 
		    entry->negation ? "! " : "", option) ? 0 : -EINVAL;
		break;
	case AMSELWALL_ENTRY_TO_DEST:
		ret = asprintf(&entry->option, "-d %s%s", 
		    entry->negation ? "! " : "", option) ? 0 : -EINVAL;
		break;
	case AMSELWALL_ENTRY_TO_IFACE:
		ret = asprintf(&entry->option, "-o %s%s", 
		    entry->negation ? "! " : "", option) ? 0 : -EINVAL;
		break;
	case AMSELWALL_ENTRY_TO_LOGICAL_IFACE:
		ret = asprintf(&entry->option, "--logical-out %s%s", 
		    entry->negation ? "! " : "", option) ? 0 : -EINVAL;
		break;
	case AMSELWALL_ENTRY_TO_PORT:
		ret = asprintf(&entry->option, "--dport %s%s", 
		    entry->negation ? "! " : "", option) ? 0 : -EINVAL;
		break;
	default:
		ret = -ENOTSUP;
	}

	if(ret != 0)
		return(ret);

	if(entry->option == NULL) {
		free(entry);
		return(-ENOMEM);
	}

	entry->type = type;
	entry->rule = rule;

	TAILQ_INSERT_TAIL(&rule->h_head, entry, entries);

	if(entry_ptr != NULL)
		*entry_ptr = entry;

	return(0);
}

/*
 * Generate commandline output
 */
int
amwall_generate(struct amselwall_table *table)
{
	struct amselwall_chain *chain;
	struct amselwall_rule *rule;
	struct amselwall_entry *entry;
	char *prog;

	switch(table->type) {
	case AMSELWALL_TABLE_IPV4:
		prog = AMSELWALL_PROG_IPTABLES;
		break;
	case AMSELWALL_TABLE_IPV6:
		prog = AMSELWALL_PROG_IP6TABLES;
		break;
	case AMSELWALL_TABLE_BRIDGE:
		prog = AMSELWALL_PROG_EBTABLES;
		break;
	case AMSELWALL_TABLE_IPVS:
		return(amwall_ipvs_generate(table));
	default:
		return(-EINVAL);
	}

	if(!(conf->flags & AMSELWALL_OPT_KEEP) || 
	    conf->flags & AMSELWALL_OPT_FLUSH) {
		/*
		 * Flush all rules and delete all user-defined chains
		 */
		printf("%s -t %s -F\n", prog, table->name);
		printf("%s -t %s -X\n", prog, table->name);
	}

	TAILQ_FOREACH(chain, &table->h_head, entries) {
		if(chain->type == AMSELWALL_CHAIN_CUSTOM)
			printf("%s -t %s -N %s\n", prog, table->name, chain->name);
		
		if(conf->flags & AMSELWALL_OPT_FLUSH) {
			if(chain->type > AMSELWALL_CHAIN_CUSTOM &&
			    chain->type < AMSELWALL_CHAIN_PREROUTING)
				printf("%s -P %s ACCEPT\n", prog, chain->name);
			continue;
		}

		TAILQ_FOREACH(rule, &chain->h_head, entries) {
			switch(rule->type) {
			case AMSELWALL_RULE_ADD:
				printf("%s -t %s -A %s ", 
				    prog, table->name, chain->name);
				
				TAILQ_FOREACH(entry, &rule->h_head, entries)
				    printf("%s ", entry->option);
				
				printf("-j %s", rule->target);

				if(rule->options == NULL)
					printf("\n");
				else
					printf(" %s\n", rule->options);
				break;
			case AMSELWALL_RULE_POLICY:
				printf("%s -t %s -P %s %s\n",
				    prog, table->name, chain->name, rule->target);
				break;
			}
		}
	}

	return(0);
}

/*
 * Cleanup parse tree
 */
void
amwall_cleanup(void)
{
	struct amselwall_table *table, *next_table;
	struct amselwall_chain *chain, *next_chain;
	struct amselwall_rule *rule, *next_rule;
	struct amselwall_entry *entry, *next_entry;

	for(table = TAILQ_FIRST(&conf->h_head); 
	    table != NULL; table = next_table) {
		next_table = TAILQ_NEXT(table, entries);
		TAILQ_REMOVE(&conf->h_head, 
		    table, entries);
		
		for(chain = TAILQ_FIRST(&table->h_head); 
		    chain != NULL; chain = next_chain) {
			next_chain = TAILQ_NEXT(chain, entries);
			TAILQ_REMOVE(&table->h_head, 
			    chain, entries);

			for(rule = TAILQ_FIRST(&chain->h_head); 
			    rule != NULL; rule = next_rule) {
				next_rule = TAILQ_NEXT(rule, entries);
				TAILQ_REMOVE(&chain->h_head, 
				    rule, entries);
				
				for(entry = TAILQ_FIRST(&rule->h_head); 
				    entry != NULL; entry = next_entry) {
					next_entry = TAILQ_NEXT(entry, entries);
					TAILQ_REMOVE(&rule->h_head, 
					    entry, entries);

					if(entry->option)
						free(entry->option);
					free(entry);
				}

				if(rule->target)
					free(rule->target);
				if(rule->options)
					free(rule->options);
				free(rule);
			}

			free(chain->name);
			free(chain);
		}

		if(table->type == AMSELWALL_TABLE_IPVS)
			amwall_ipvs_cleanup(table);

		free(table->name);
		free(table);
	}
}

