/* -*- knf -*- */
/* $AMSEL: tables.h,v 1.2 2004/06/29 15:59:13 reyk Exp $ */

/*
 * Copyright (c) 2004 .vantronix | secure systems GmbH
 * Copyright (c) 2004 SSV Embedded Systems GmbH
 *
 * All rights reserved.
 *
 * Written by Reyk Floeter <reyk@vantronix.net>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  *
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#ifndef _AMSEL_WALL_TABLES_HEADER
#define _AMSEL_WALL_TABLES_HEADER

#define AMSELWALL_PROG_IPTABLES "/sbin/iptables"
#define AMSELWALL_PROG_IP6TABLES "/sbin/ip6tables"
#define AMSELWALL_PROG_EBTABLES "/sbin/ebtables"

struct amselwall_entry;
struct amselwall_rule;
struct amselwall_chain;
struct amselwall_table;

enum amselwall_entry_type {
	AMSELWALL_ENTRY_MISC = 0,
	AMSELWALL_ENTRY_FRAGMENT,
	AMSELWALL_ENTRY_CONDITION_PROTO,
	AMSELWALL_ENTRY_FROM_SOURCE,
	AMSELWALL_ENTRY_FROM_IFACE,
	AMSELWALL_ENTRY_FROM_LOGICAL_IFACE,
	AMSELWALL_ENTRY_FROM_PORT,
	AMSELWALL_ENTRY_TO_DEST,
	AMSELWALL_ENTRY_TO_IFACE,
	AMSELWALL_ENTRY_TO_LOGICAL_IFACE,
	AMSELWALL_ENTRY_TO_PORT,
	AMSELWALL_ENTRY_MATCH,
	AMSELWALL_ENTRY_OPTIONS,
};

struct amselwall_entry {
	char *option;
	int negation;
	enum amselwall_entry_type type;
	struct amselwall_rule *rule;
	TAILQ_ENTRY(amselwall_entry) entries;
};

enum amselwall_rule_type {
	AMSELWALL_RULE_ADD = 0,
	AMSELWALL_RULE_POLICY,
};

struct amselwall_rule {
	char *target, *options;
	enum amselwall_rule_type type;
	struct amselwall_chain *chain;
	TAILQ_ENTRY(amselwall_rule) entries;
	TAILQ_HEAD(amselwall_entry_head, amselwall_entry) h_head;
};

enum amselwall_chain_type {
	AMSELWALL_CHAIN_CUSTOM = 0,
	AMSELWALL_CHAIN_INPUT = 1,
	AMSELWALL_CHAIN_FORWARD = 2,
	AMSELWALL_CHAIN_OUTPUT = 3,
	AMSELWALL_CHAIN_PREROUTING = 4,
	AMSELWALL_CHAIN_POSTROUTING = 5,
	AMSELWALL_CHAIN_BROUTING = 6, /* ebtables */
};

struct amselwall_chain {
	char *name;
	enum amselwall_chain_type type;
	struct amselwall_table *table;
	TAILQ_ENTRY(amselwall_chain) entries;
	TAILQ_HEAD(amselwall_rule_head, amselwall_rule) h_head;
};

enum amselwall_table_type {
	AMSELWALL_TABLE_IPV4 = 0,
	AMSELWALL_TABLE_IPV6,
	AMSELWALL_TABLE_BRIDGE,
	AMSELWALL_TABLE_IPVS,
};

struct amselwall_table {
	char *name;
	enum amselwall_table_type type;
	TAILQ_ENTRY(amselwall_table) entries;
	TAILQ_HEAD(amselwall_chain_head, amselwall_chain) h_head;
	void *data;
};

extern int amwall_add_table(char *, enum amselwall_table_type,
    struct amselwall_table **);
extern int amwall_add_chain(struct amselwall_table *, char *, 
    enum amselwall_chain_type, struct amselwall_chain **);
extern int amwall_add_rule(struct amselwall_chain *, char *, 
    enum amselwall_table_type, struct amselwall_rule **);

extern int amwall_add_rule_options(struct amselwall_rule *, char *);
extern int amwall_add_entry(struct amselwall_rule *, char *, int,
    enum amselwall_entry_type, struct amselwall_entry **);
extern int amwall_generate(struct amselwall_table *);
extern void amwall_cleanup(void);

#endif /* _AMSEL_WALL_TABLES_HEADER */
